/*
 * kgraft_patch_bsc1097108
 *
 * Fix for CVE-2018-10853, bsc#1097108
 *
 *  Upstream commits:
 *  79367a657439 ("KVM: x86: introduce linear_{read,write}_system")
 *  ce14e868a54e ("KVM: x86: pass kvm_vcpu to kvm_read_guest_virt and
 *                 kvm_write_guest_virt_system")
 *  3c9fa24ca7c9 ("kvm: x86: use correct privilege level for
 *                 sgdt/sidt/fxsave/fxrstor access")
 *
 *  SLE12 commit:
 *  none yet
 *
 *  SLE12-SP1 commit
 *  none yet
 *
 *  SLE12-SP2 commit:
 *  none yet
 *
 *  SLE12-SP3 commit:
 *  06238a187fb7aafb7088fc617cb0a0efd6259d80 (stable 4.4.138)
 *
 *  SLE15 commit:
 *  none yet
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */


#if IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/kvm_host.h>
#include <asm/kvm_emulate.h>
#include <asm/desc_defs.h>
#include "kgr_patch_bsc1097108.h"

#if !IS_MODULE(CONFIG_KVM)
#error "Live patch supports only CONFIG_KVM=m"
#endif

#define KGR_PATCHED_MODULE "kvm"


static struct kvm_x86_ops **kgr_kvm_x86_ops;

static int (*kgr_kvm_read_guest_virt_system)(struct x86_emulate_ctxt *ctxt,
					     gva_t addr, void *val,
					     unsigned int bytes,
					     struct x86_exception *exception);
static int (*kgr_kvm_write_guest_virt_system)(struct x86_emulate_ctxt *ctxt,
					      gva_t addr, void *val,
					      unsigned int bytes,
					      struct x86_exception *exception);
static int (*kgr_kvm_read_guest_virt_helper)(gva_t addr, void *val,
					     unsigned int bytes,
					     struct kvm_vcpu *vcpu, u32 access,
					     struct x86_exception *exception);
static int (*kgr_kvm_write_guest)(struct kvm *kvm, gpa_t gpa, const void *data,
				  unsigned long len);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "kvm:kvm_x86_ops", (void *)&kgr_kvm_x86_ops },
	{ "kvm:kvm_read_guest_virt_system",
				(void *)&kgr_kvm_read_guest_virt_system },
	{ "kvm:kvm_write_guest_virt_system",
				(void *)&kgr_kvm_write_guest_virt_system },
	{ "kvm:kvm_read_guest_virt_helper",
				(void *)&kgr_kvm_read_guest_virt_helper },
	{ "kvm:kvm_write_guest", (void *)&kgr_kvm_write_guest },
};



/* from arch/x86/kvm/x86.c */
#define kgr_emul_to_vcpu(ctxt)					\
	container_of(ctxt, struct kvm_vcpu, arch.emulate_ctxt)


/* from arch/x86/kvm/mmu.h */
#define KGR_PFERR_WRITE_BIT 1
#define KGR_PFERR_USER_BIT 2
#define KGR_PFERR_WRITE_MASK (1U << KGR_PFERR_WRITE_BIT)
#define KGR_PFERR_USER_MASK (1U << KGR_PFERR_USER_BIT)


/* from arch/x86/kvm/emulate.c */
#define KGR_Aligned     ((u64)1 << 41)  /* Explicitly aligned (e.g. MOVDQA) */
#define KGR_Unaligned   ((u64)1 << 42)  /* Explicitly unaligned (e.g. MOVDQU) */
#define KGR_Avx         ((u64)1 << 43)  /* Advanced Vector Extensions */

/* inlined */
static u32 kgr_desc_limit_scaled(struct desc_struct *desc)
{
	u32 limit = get_desc_limit(desc);

	return desc->g ? (limit << 12) | 0xfff : limit;
}

/* inlined */
static unsigned long kgr_seg_base(struct x86_emulate_ctxt *ctxt, int seg)
{
	if (ctxt->mode == X86EMUL_MODE_PROT64 && seg < VCPU_SREG_FS)
		return 0;

	return ctxt->ops->get_cached_segment_base(ctxt, seg);
}

/* inlined */
int kgr_emulate_exception(struct x86_emulate_ctxt *ctxt, int vec,
			  u32 error, bool valid)
{
	ctxt->exception.vector = vec;
	ctxt->exception.error_code = error;
	ctxt->exception.error_code_valid = valid;
	return X86EMUL_PROPAGATE_FAULT;
}

/* inlined */
static int kgr_emulate_gp(struct x86_emulate_ctxt *ctxt, int err)
{
	return kgr_emulate_exception(ctxt, GP_VECTOR, err, true);
}

/* inlined */
static int kgr_emulate_ss(struct x86_emulate_ctxt *ctxt, int err)
{
	return kgr_emulate_exception(ctxt, SS_VECTOR, err, true);
}

/* inlined */
static bool kgr_insn_aligned(struct x86_emulate_ctxt *ctxt, unsigned size)
{
	if (likely(size < 16))
		return false;

	if (ctxt->d & KGR_Aligned)
		return true;
	else if (ctxt->d & KGR_Unaligned)
		return false;
	else if (ctxt->d & KGR_Avx)
		return false;
	else
		return true;
}

/* optimized */
static int kgr__linearize(struct x86_emulate_ctxt *ctxt,
			  struct segmented_address addr,
			  unsigned size, bool write, bool fetch,
			  ulong *linear)
{
	struct desc_struct desc;
	bool usable;
	ulong la;
	u32 lim;
	u16 sel;
	unsigned cpl;

	la = kgr_seg_base(ctxt, addr.seg) + addr.ea;
	switch (ctxt->mode) {
	case X86EMUL_MODE_PROT64:
		if (is_noncanonical_address(la))
			return kgr_emulate_gp(ctxt, 0);
		break;
	default:
		usable = ctxt->ops->get_segment(ctxt, &sel, &desc, NULL,
						addr.seg);
		if (!usable)
			goto bad;
		/* code segment in protected mode or read-only data segment */
		if ((((ctxt->mode != X86EMUL_MODE_REAL) && (desc.type & 8))
					|| !(desc.type & 2)) && write)
			goto bad;
		/* unreadable code segment */
		if (!fetch && (desc.type & 8) && !(desc.type & 2))
			goto bad;
		lim = kgr_desc_limit_scaled(&desc);
		if ((desc.type & 8) || !(desc.type & 4)) {
			/* expand-up segment */
			if (addr.ea > lim || (u32)(addr.ea + size - 1) > lim)
				goto bad;
		} else {
			/* expand-down segment */
			if (addr.ea <= lim || (u32)(addr.ea + size - 1) <= lim)
				goto bad;
			lim = desc.d ? 0xffffffff : 0xffff;
			if (addr.ea > lim || (u32)(addr.ea + size - 1) > lim)
				goto bad;
		}
		cpl = ctxt->ops->cpl(ctxt);
		if (!(desc.type & 8)) {
			/* data segment */
			if (cpl > desc.dpl)
				goto bad;
		} else if ((desc.type & 8) && !(desc.type & 4)) {
			/* nonconforming code segment */
			if (cpl != desc.dpl)
				goto bad;
		} else if ((desc.type & 8) && (desc.type & 4)) {
			/* conforming code segment */
			if (cpl < desc.dpl)
				goto bad;
		}
		break;
	}
	if (fetch ? ctxt->mode != X86EMUL_MODE_PROT64 : ctxt->ad_bytes != 8)
		la &= (u32)-1;
	if (kgr_insn_aligned(ctxt, size) && ((la & (size - 1)) != 0))
		return kgr_emulate_gp(ctxt, 0);
	*linear = la;
	return X86EMUL_CONTINUE;
bad:
	if (addr.seg == VCPU_SREG_SS)
		return kgr_emulate_ss(ctxt, sel);
	else
		return kgr_emulate_gp(ctxt, sel);
}

/* optimized */
static int kgr_linearize(struct x86_emulate_ctxt *ctxt,
			 struct segmented_address addr,
			 unsigned size, bool write,
			 ulong *linear)
{
	return kgr__linearize(ctxt, addr, size, write, false, linear);
}



/* New */
static int kgr_emulator_read_std(struct x86_emulate_ctxt *ctxt,
				 gva_t addr, void *val, unsigned int bytes,
				 struct x86_exception *exception, bool system)
{
	struct kvm_vcpu *vcpu = kgr_emul_to_vcpu(ctxt);
	u32 access = 0;

	if (!system && (*kgr_kvm_x86_ops)->get_cpl(vcpu) == 3)
		access |= KGR_PFERR_USER_MASK;

	return kgr_kvm_read_guest_virt_helper(addr, val, bytes, vcpu, access,
					  exception);
}

/* New */
static int kgr_kvm_write_guest_virt_helper(gva_t addr, void *val,
					   unsigned int bytes,
					   struct kvm_vcpu *vcpu, u32 access,
					   struct x86_exception *exception)
{
	void *data = val;
	int r = X86EMUL_CONTINUE;

	while (bytes) {
		gpa_t gpa =  vcpu->arch.walk_mmu->gva_to_gpa(vcpu, addr,
							     access,
							     exception);
		unsigned offset = addr & (PAGE_SIZE-1);
		unsigned towrite = min(bytes, (unsigned)PAGE_SIZE - offset);
		int ret;

		if (gpa == UNMAPPED_GVA)
			return X86EMUL_PROPAGATE_FAULT;
		ret = kgr_kvm_write_guest(vcpu->kvm, gpa, data, towrite);
		if (ret < 0) {
			r = X86EMUL_IO_NEEDED;
			goto out;
		}

		bytes -= towrite;
		data += towrite;
		addr += towrite;
	}
out:
	return r;
}

/* New */
static int kgr_emulator_write_std(struct x86_emulate_ctxt *ctxt, gva_t addr,
				  void *val, unsigned int bytes,
				  struct x86_exception *exception, bool system)
{
	struct kvm_vcpu *vcpu = kgr_emul_to_vcpu(ctxt);
	u32 access = KGR_PFERR_WRITE_MASK;

	if (!system && (*kgr_kvm_x86_ops)->get_cpl(vcpu) == 3)
		access |= KGR_PFERR_USER_MASK;

	return kgr_kvm_write_guest_virt_helper(addr, val, bytes, vcpu,
					       access, exception);
}

/* patched, optimized */
static int kgr_segmented_read_std(struct x86_emulate_ctxt *ctxt,
				  struct segmented_address addr,
				  void *data,
				  unsigned size)
{
	int rc;
	ulong linear;

	rc = kgr_linearize(ctxt, addr, size, false, &linear);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	/*
	 * Fix CVE-2018-10853
	 *  +5 lines
	 *
	 * Note that ->read_std should always equal
	 * kvm_read_guest_virt_system and that the check below is
	 * only there as a safety measure.
	 */
	if (likely(ctxt->ops->read_std == kgr_kvm_read_guest_virt_system)) {
		return kgr_emulator_read_std(ctxt, linear, data, size,
					     &ctxt->exception, false);
	}

	return ctxt->ops->read_std(ctxt, linear, data, size, &ctxt->exception);
}

/* patched, inlined */
static int kgr_segmented_write_std(struct x86_emulate_ctxt *ctxt,
				   struct segmented_address addr,
				   void *data,
				   unsigned int size)
{
	int rc;
	ulong linear;

	rc = kgr_linearize(ctxt, addr, size, true, &linear);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	/*
	 * Fix CVE-2018-10853
	 *  +5 lines
	 *
	 * Note that ->write_std should always equal
	 * kvm_write_guest_virt_system and that the check below is
	 * only there as a safety measure.
	 */
	if (likely(ctxt->ops->write_std == kgr_kvm_write_guest_virt_system)) {
		return kgr_emulator_write_std(ctxt, linear, data, size,
					      &ctxt->exception, false);
	}

	return ctxt->ops->write_std(ctxt, linear, data, size, &ctxt->exception);
}


/* patched, optimized, calls segmented_read_std() */
static int kgr_read_descriptor(struct x86_emulate_ctxt *ctxt,
			       struct segmented_address addr,
			       u16 *size, unsigned long *address, int op_bytes)
{
	int rc;

	if (op_bytes == 2)
		op_bytes = 3;
	*address = 0;
	rc = kgr_segmented_read_std(ctxt, addr, size, 2);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	addr.ea += 2;
	rc = kgr_segmented_read_std(ctxt, addr, address, op_bytes);
	return rc;
}

/* patched, calls read_descriptor() */
int kgr_em_lgdt(struct x86_emulate_ctxt *ctxt)
{
	struct desc_ptr desc_ptr;
	int rc;

	if (ctxt->mode == X86EMUL_MODE_PROT64)
		ctxt->op_bytes = 8;
	rc = kgr_read_descriptor(ctxt, ctxt->src.addr.mem,
				 &desc_ptr.size, &desc_ptr.address,
				 ctxt->op_bytes);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	ctxt->ops->set_gdt(ctxt, &desc_ptr);
	/* Disable writeback. */
	ctxt->dst.type = OP_NONE;
	return X86EMUL_CONTINUE;
}

/* patched, calls read_descriptor() */
int kgr_em_lidt(struct x86_emulate_ctxt *ctxt)
{
	struct desc_ptr desc_ptr;
	int rc;

	if (ctxt->mode == X86EMUL_MODE_PROT64)
		ctxt->op_bytes = 8;
	rc = kgr_read_descriptor(ctxt, ctxt->src.addr.mem,
				 &desc_ptr.size, &desc_ptr.address,
				 ctxt->op_bytes);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	ctxt->ops->set_idt(ctxt, &desc_ptr);
	/* Disable writeback. */
	ctxt->dst.type = OP_NONE;
	return X86EMUL_CONTINUE;
}


/* patched, calls segmented_write_std() */
int kgr_emulate_store_desc_ptr(struct x86_emulate_ctxt *ctxt,
			       void (*get)(struct x86_emulate_ctxt *ctxt,
					   struct desc_ptr *ptr))
{
	struct desc_ptr desc_ptr;

	if (ctxt->mode == X86EMUL_MODE_PROT64)
		ctxt->op_bytes = 8;
	get(ctxt, &desc_ptr);
	if (ctxt->op_bytes == 2) {
		ctxt->op_bytes = 4;
		desc_ptr.address &= 0x00ffffff;
	}
	/* Disable writeback. */
	ctxt->dst.type = OP_NONE;
	return kgr_segmented_write_std(ctxt, ctxt->dst.addr.mem,
				       &desc_ptr, 2 + ctxt->op_bytes);
}



static int kgr_patch_bsc1097108_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1097108_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1097108_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1097108_module_nb = {
	.notifier_call = kgr_patch_bsc1097108_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1097108_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1097108_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1097108_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1097108_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1097108_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM) */
